{ config, lib, pkgs, ... }:

let
  svc_domain = "stalwart.${config.mine.shared.settings.domain}";
  svc_domain_mail = "mail.${config.mine.shared.settings.domain}";
  # TODO(eyJhb): in theory these domains are not used, they're just
  # nice to have.. maybe.
  svc_domain_smtp = "smtp.${config.mine.shared.settings.domain}";
  svc_domain_imap = "imap.${config.mine.shared.settings.domain}";

  ports = {
    smtp = 25;
    submissions = 465;
    imaptls = 993;

    http_management = 7377;
  };

  # place data into own zfs dataset
  stateDir = config.mine.zfsMounts."rpool/safe/svcs/stalwart";

  # user/group
  stalwart_user = config.users.users.stalwart-mail.name;
  stalwart_group = config.users.groups.stalwart-mail.name;

  certLocation = config.security.acme.certs."${svc_domain_mail}".directory;
in {
  services.stalwart-mail = {
    enable = true;
    openFirewall = true;

    settings = {
      lookup.default.hostname = svc_domain_mail;

      store.db.path = "${stateDir}/db";

      directory.ldap = {
        type = "ldap";
        url = config.mine.shared.settings.ldap.url;
        base-dn = config.mine.shared.settings.ldap.dc;
        tls.enable = false;

        bind = {
          dn = config.mine.shared.settings.ldap.bind_dn;
          secret = "%{file:${config.age.secrets.lldap-bind-user-pass.path}}%";

          # TODO: I'm not sure if this should be lookup, or template
          # LLDAPs config is out of date, and there is a closed issue here
          # https://github.com/stalwartlabs/stalwart/issues/1730
          # this was changed in v.0.12.4, changelog here
          # https://github.com/stalwartlabs/stalwart/releases/tag/v0.12.4
          auth = {
            # method = "template";
            # dn = "cn={username},${config.mine.shared.settings.ldap.search_base}";
            method = "lookup";
            dn = "cn=?,${config.mine.shared.settings.ldap.search_base}";
            search = true;
          };
        };

        filter = let
          _mkFilter = attrs: ph: config.mine.shared.lib.ldap.mkFilter (lconfig: llib:
            llib.mkAnd [
              (llib.mkOr [
                (llib.mkGroup lconfig.groups.member)
                (llib.mkGroup lconfig.groups.system_mail)
              ])
              (llib.mkOr (lib.forEach attrs (v: llib.mkSearch v ph)))
            ]
          );

          attrs = config.mine.shared.settings.ldap.attr // { emailAlias = "mailAlias"; emailList = "mailList"; };
        in {
          name = _mkFilter [ attrs.uid ] "?";
          email = _mkFilter [ attrs.membermail ] "?";
        };

        attributes = config.mine.shared.lib.ldap.mkScope (lconfig: llib: {
          name = lconfig.attr.uid;
          # name = lconfig.attr.member_mail;
          description = lconfig.attr.firstname;
          email = lconfig.attr.membermail;
          quota = lconfig.attr.membermaildiskquota;
          class = "objectClass";
          groups = "memberOf";
          # we dont have access to this in lldap, and
          # therefore we use secret-changed instead
          # secret = lconfig.attr.stalwart_secret;
          secret = lconfig.attr.creationdate;
          # TODO(eyJhb): remove once LLDAP gets
          # plugin support, so we can make a plugin
          # that updates a attribute on password updates
          # https://github.com/lldap/lldap/pull/1119
          secret-changed = lconfig.attr.creationdate;
        });
      };

      storage.directory = "ldap";

      # listeners
      server.listener = {
        smtp = { bind = [ "[::]:${builtins.toString ports.smtp}"]; protocol = "smtp"; };
        submissions = { bind = [ "[::]:${builtins.toString ports.submissions}"]; protocol = "smtp"; tls.implicit = true; };
        imaptls = { bind = [ "[::]:${builtins.toString ports.imaptls}"]; protocol = "imap"; tls.implicit = true; };

        management = { bind = [ "[::]:${builtins.toString ports.http_management}" ]; protocol = "http"; };
      };

      certificate.domain = {
        cert = "%{file:${certLocation + "/cert.pem"}}%";
        private-key = "%{file:${certLocation + "/key.pem"}}%";
      };

      server.tls = {
        certificate = "domain";
        enable = true;
        implicit = true;
      };

      # authentication
      authentication.fallback-admin = {
        user = "admin";
        secret = "%{file:${config.age.secrets.stalwart-admin-fallback-password.path}}%";
      };
    };
  };

  # ensure that the configured domain is added to stalwart
  systemd.services.stalwart-mail.serviceConfig.ExecStartPost = let
    stalwart-cli = config.services.stalwart-mail.package + "/bin/stalwart-cli";
  in pkgs.writeShellScript "stalwart-setup" ''
    # TODO(eyjhb): at some point this should be changed
    sleep 10

    export URL="http://127.0.0.1:${builtins.toString ports.http_management}"
    export CREDENTIALS="$(cat ${config.age.secrets.stalwart-admin-fallback-password.path})"

    if ${stalwart-cli} domain list | grep --quiet -- '${config.mine.shared.settings.domain}'; then
      echo "Domain is already created, doing nothing."
    else
      echo "Domain was not created, creating it now."
      ${stalwart-cli} domain create '${config.mine.shared.settings.domain}'
    fi

    echo "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!IMPORTANT!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
    echo "Remember to setup the correct DNS records. Run the following command to get them:"
    echo "URL=\"$URL\" stalwart-cli domain dns-records ${config.mine.shared.settings.domain}"
    echo "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!IMPORTANT!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
  '';

  # setup so that stalwart can access and write to the directory
  systemd.services.stalwart-mail.serviceConfig.ReadWritePaths = [ stateDir ];
  systemd.tmpfiles.rules = [
    "Z ${stateDir} 0700 ${stalwart_user} ${stalwart_group} -"
  ];

  # setup certs
  services.nginx.virtualHosts."${svc_domain_mail}" = {
    forceSSL = true;
    enableACME = true;
    serverAliases = [ svc_domain_smtp svc_domain_imap ];
    root = pkgs.writeTextDir "index.html" "Nothing.";
  };

  services.nginx.virtualHosts."${svc_domain}" = {
    forceSSL = true;
    enableACME = true;
    locations."/".proxyPass = "http://localhost:${builtins.toString ports.http_management}";
  };

  # need to change group to stalwart-mail for cert + add nginx to stalwart-mail group to do HTTP ACME
  users.users.nginx.extraGroups = [ stalwart_group ];
  security.acme.certs."${svc_domain_mail}" = {
    group = stalwart_group; 
    reloadServices = [ config.systemd.services.stalwart-mail.name ];
  };

  # setup secrets for stalwart
  # setup access to ldap bind user credential
  users.groups.secrets-lldap-bind-user-pass.members = [ stalwart_user ];
  age.secrets.stalwart-admin-fallback-password.owner = stalwart_user;

  mine.shared.settings.mail = {
    domain = svc_domain;
    domain_mail = svc_domain_mail;
    domain_smtp = svc_domain_smtp;
    domain_imap = svc_domain_imap;

    ports = ports;
  };
}
